"""
_WindowsUSBDetector: This platform-specific implementation of the _USBDetectorBase class is designed for Windows
systems. It provides the necessary functionality to detect USB devices connected to a Windows system and monitor
changes in their connections. The class utilizes the WMI (Windows Management Instrumentation) library to interact with
the Windows system's device management subsystem.

Author: Eric-Canas
Date: 27-03-2023
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""

from __future__ import annotations
import re
from warnings import warn

from ..attributes import DEVTYPE
from ._constants import _DEVICE_ID, _LINUX_TO_WINDOWS_ATTRIBUTES, _SECONDS_BETWEEN_CHECKS, \
    _WINDOWS_USB_REGEX_ATTRIBUTES, \
    _WINDOWS_NON_USB_DEVICES_IDS, _WINDOWS_USB_QUERY, _WINDOWS_TO_LOWERCASE_ATTRIBUTES, \
    _WINDOWS_REGEX_ATTRIBUTES_BY_DRIVER

from ._usb_detector_base import _USBDetectorBase


class _WindowsUSBDetector(_USBDetectorBase):
    def __init__(self, filter_devices: list[dict[str, str]] | tuple[dict[str, str]] | None = None):
        self._wmi_interface = None
        # CONSTANTS
        super(_WindowsUSBDetector, self).__init__(filter_devices=filter_devices)

    def get_available_devices(self) -> dict[str, dict[str, str]]:
        """
        Returns a dictionary of the currently available devices, where the key is the device ID and the value is a
        dictionary of the device's information.
        :return: dict[str, dict[str, str]]. The key is the device ID, the value is a dictionary of the device's
                information.
        """
        if self._wmi_interface is None:
            self._wmi_interface = self.__create_wmi_interface()
        devices = self._wmi_interface.query(_WINDOWS_USB_QUERY)
        devices = {getattr(device, _DEVICE_ID): {new_name: getattr(device, attribute)
                                                 for new_name, attribute in _LINUX_TO_WINDOWS_ATTRIBUTES.items()}
                    for device in devices}
        devices = self.__filter_devices(devices=devices)
        devices = self.__finetune_incompatible_attributes(devices=devices)
        if self.filter_devices is not None:
            devices = self._apply_devices_filter(devices=devices)
        return devices

    def _monitor_changes(self, on_connect: callable | None = None, on_disconnect: callable | None = None,
                         check_every_seconds: int | float = _SECONDS_BETWEEN_CHECKS) -> None:
        """
        Monitors the USB devices. This function should ALWAYS be called from a background thread.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive two
                arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to receive
                two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])
        :param check_every_seconds: int | float. The number of seconds to wait between each check for changes in the
                USB devices. Defaults to 0.5 seconds.
        """
        # If running this in a background thread, we MUST create the WMI interface inside the thread.
        self._wmi_interface = self.__create_wmi_interface()
        super(_WindowsUSBDetector, self)._monitor_changes(on_connect=on_connect, on_disconnect=on_disconnect,
                                                         check_every_seconds=check_every_seconds)

    def __filter_devices(self, devices: dict[str, dict[str, tuple[str]|str]]) -> dict[str, dict[str, tuple[str]|str]]:
        """
        Filters the devices to only include the ones that are USB devices.
        :param devices: dict[str, dict[str, str]]. The dictionary of devices to filter.
        :return: dict[str, dict[str, str]]. The filtered devices.
        """
        return {device_id: device_info for device_id, device_info in devices.items()
                if not any(substring in device_info[DEVTYPE] for substring in _WINDOWS_NON_USB_DEVICES_IDS)}

    def __finetune_incompatible_attributes(self, devices: dict[str, dict[str | tuple[str, ...]]]) -> dict[str, dict[str, str]]:
        """
        Transforms some attributes to be more similar to the Linux attributes.
        :param devices: dict[str, str|tuple[str,...]]. The dictionary of devices to transform.
        :return: dict[str, dict[str, str]]. The transformed devices.
        """
        for device_id, device_info in devices.items():
            driver_type = self.__get_driver_type_from_device_id(device_id=device_id)
            new_attributes = {attribute: self.__apply_regex(value=device_id, regex=regex)#device_info[attribute], regex)
                              for attribute, regex in _WINDOWS_REGEX_ATTRIBUTES_BY_DRIVER[driver_type].items()
                              if attribute in device_info}
            device_info.update(new_attributes)
            new_attributes = {attr: device_info[attr].upper() for attr in _WINDOWS_TO_LOWERCASE_ATTRIBUTES}
            device_info.update(new_attributes)
        return devices

    def __apply_regex(self, value: tuple[str] | str, regex: str) -> str:
        """
        Apply the regex to a value, taking into account if it is a tuple or not.
        :param value: tuple[str] | str. The value to apply the regex to.
        :param regex: str. The regex to apply.
        :return: str. The value after applying the regex.
        """
        if isinstance(value, str):
            value = (value,)
        values_found = []
        for value in value:
            match = re.search(regex, value)
            if match is not None:
                values_found.append(match.group(1))
        # If no value was found, return the original value
        if len(values_found) == 0:
            warn(f"Could not find a value for the regex '{regex}' in the value '{value}'")
            return value
        # Otherwise, return check there are no inconsistencies and return the value
        assert all(value == values_found[0] for value in values_found), "The values found are not all the same"
        return values_found[0]

    def __get_driver_type_from_device_id(self, device_id: str) -> str:
        """
        Returns the driver type from the device ID.
        :param device_id: str. The device ID.
        :return: str. The driver type.
        """
        device_splitted_info = device_id.split('\\')
        assert len(device_splitted_info) >= 1, f"The device ID '{device_id}' is not valid"
        driver_type = device_splitted_info[0]
        assert driver_type in _WINDOWS_REGEX_ATTRIBUTES_BY_DRIVER, f"The driver type '{driver_type}' is not supported " \
                                                                   f"yet, please create an issue in github"
        return driver_type

    def __create_wmi_interface(self):
        from pythoncom import CoInitialize
        CoInitialize()
        import wmi
        # If running this in a background thread, we MUST create the WMI interface inside the thread.
        return wmi.WMI()


"""
device_info dictionary attributes. Same attributes as Linux udev (independently of the OS where the library is running).
All these attributes are the keys of the device_info dictionary returned by the `get_available_devices` method.
The expected values of these attributes are all strings, except for the `ID_USB_INTERFACES` attribute, which is a tuple.

Author: Eric-Canas
Date: 26-03-2023
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""

ID_VENDOR_ID = 'ID_VENDOR_ID'
ID_VENDOR = 'ID_VENDOR'
ID_MODEL = 'ID_MODEL'
ID_MODEL_ID = 'ID_MODEL_ID'
ID_SERIAL = 'ID_SERIAL'
ID_USB_INTERFACES = 'ID_USB_INTERFACES'
ID_REVISION = 'ID_REVISION'

ID_USB_CLASS_FROM_DATABASE = 'ID_USB_CLASS_FROM_DATABASE'
ID_VENDOR_FROM_DATABASE = 'ID_VENDOR_FROM_DATABASE'
ID_MODEL_FROM_DATABASE = 'ID_MODEL_FROM_DATABASE'

DEVNAME = 'DEVNAME'
DEVTYPE = 'DEVTYPE'

DEVICE_ATTRIBUTES = (ID_MODEL_ID, ID_MODEL, ID_MODEL_FROM_DATABASE, ID_VENDOR, ID_VENDOR_ID, ID_VENDOR_FROM_DATABASE,
                     ID_USB_INTERFACES, ID_USB_CLASS_FROM_DATABASE, DEVNAME, DEVTYPE, ID_SERIAL)



"""
_DarwinUSBDetector: This platform-specific implementation of the _USBDetectorBase class is designed for MacOS systems.
It provides the necessary functionality to detect USB devices connected to a MacOS system and monitor changes in their
connections. The class utilizes the ioreg command to interact with the MacOS system's device management subsystem.

Author: Eric-Canas
Date: 01-06-2024
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""

from __future__ import annotations
import re
import subprocess
from warnings import warn

from ..attributes import DEVTYPE, ID_VENDOR_ID, DEVNAME, DEVICE_ATTRIBUTES
from ._constants import _DARWIN_TO_LINUX_ATTRIBUTES, _DARWIN_REGEX_ATTRIBUTES, _SECONDS_BETWEEN_CHECKS
from ._usb_detector_base import _USBDetectorBase


class _DarwinUSBDetector(_USBDetectorBase):
    def __init__(self, filter_devices: list[dict[str, str]] | tuple[dict[str, str]] | None = None):
        super(_DarwinUSBDetector, self).__init__(filter_devices=filter_devices)

    def get_available_devices(self) -> dict[str, dict[str, str]]:
        """
        Returns a dictionary of the currently available devices, where the key is the device ID and the value is a
        dictionary of the device's information.
        :return: dict[str, dict[str, str]]. The key is the device ID, the value is a dictionary of the device's
                information.
        """
        devices_info = self.__get_usb_devices()
        if self.filter_devices is not None:
            devices_info = self._apply_devices_filter(devices=devices_info)
        return devices_info

    def _monitor_changes(self, on_connect: callable | None = None, on_disconnect: callable | None = None,
                         check_every_seconds: int | float = _SECONDS_BETWEEN_CHECKS) -> None:
        """
        Monitors the USB devices. This function should ALWAYS be called from a background thread.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive two
                arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to receive
                two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])
        :param check_every_seconds: int | float. The number of seconds to wait between each check for changes in the
                USB devices. Defaults to 0.5 seconds.
        """
        while not self._stop_thread.is_set():
            self.check_changes(on_connect=on_connect, on_disconnect=on_disconnect)
            self._stop_thread.wait(check_every_seconds)

    def __get_usb_devices(self) -> dict[str, dict[str, str]]:
        """
        Retrieves the list of USB devices using the `ioreg` command.
        :return: dict[str, dict[str, str]]. A dictionary of the device information.
        """
        try:
            ioreg_output = subprocess.check_output(['ioreg', '-p', 'IOUSB', '-w0', '-l'], text=True)
        except subprocess.CalledProcessError as e:
            warn(f"Failed to retrieve USB devices information: {e}")
            return {}

        devices_info = {}
        current_device = {}
        device_id = None

        for line in ioreg_output.splitlines():
            if "+-o" in line:
                if current_device and device_id:
                    devices_info[device_id] = current_device
                current_device = {}
                device_id_match = re.search(r"\+-o\s+(.+?)\s+<", line)
                if device_id_match:
                    device_id = device_id_match.group(1)
                else:
                    device_id = None
            else:
                for attribute, darwin_attr in _DARWIN_TO_LINUX_ATTRIBUTES.items():
                    if darwin_attr in line:
                        value = line.split('=')[-1].strip().strip("}").strip('"')
                        current_device[attribute] = value

                for attribute, regex in _DARWIN_REGEX_ATTRIBUTES.items():
                    match = re.search(regex, line)
                    if match:
                        current_device[attribute] = match.group(1)

        if current_device and device_id:
            devices_info[device_id] = current_device

        # Add the device_id as the devname to the info dict
        for device_id, device_info in devices_info.items():
            device_info[DEVNAME] = device_id

        return devices_info



"""
_USBDetectorBase: This abstract base class provides the core functionality for monitoring USB devices on different
platforms. It defines the required methods and attributes, as well as common functionality for checking device changes,
starting and stopping the monitoring process, and maintaining the state of the monitoring thread. The _USBDetectorBase
class is intended to be subclassed by platform-specific implementations to provide the necessary support for USB
device monitoring.

Author: Eric-Canas
Date: 27-03-2023
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""
from __future__ import annotations

import threading
from warnings import warn
from abc import ABC, abstractmethod

from ._constants import _SECONDS_BETWEEN_CHECKS, _THREAD_JOIN_TIMEOUT_SECONDS


class _USBDetectorBase(ABC):
    def __init__(self, filter_devices: list[dict[str, str]] | tuple[dict[str, str]] = None):
        """
        Initializes the _USBDetectorBase class.

        :param filter_devices: list[dict[str, str]] | tuple[dict[str, str]] | None. A list of dictionaries containing
        the device information to filter the devices by. If None, no filtering will be done. The dictionaries
        must contain the same keys as the dictionaries returned by the `get_available_devices` method.
        For example, if you want to only monitor devices with ID_MODEL_ID = "A2B2" or "ABCD" you could pass
        filter_devices=({"ID_MODEL_ID": "A2B2"}, {"ID_MODEL_ID": "ABCD"}).
        """
        self._thread = None
        self.filter_devices = filter_devices
        self._stop_thread = threading.Event()

        if filter_devices is not None:
            assert isinstance(filter_devices, (list, tuple)), f"filter_devices must be a list or a tuple of dicts " \
                                                              f"(or None). Got {type(filter_devices)}"
            assert all(isinstance(device, dict) for device in filter_devices), f"filter_devices must contain dicts. " \
                                                                f"Got {set(type(device) for device in filter_devices)}"

        self.on_start_devices = self.get_available_devices()
        self.last_check_devices = self.on_start_devices.copy()

    def changes_from_last_check(self, update_last_check_devices: bool = True) -> tuple[dict[str, str], dict[str, str]]:
        """
        Returns a tuple of two tuples, the first containing the device IDs of the devices that were removed, the second
        containing the device IDs of the devices that were added.
        :param update_last_check_devices: bool. Whether to update the last checked devices to the current devices
        :return: tuple[dict[str, str], dict[str, str]]. The first tuple contains the information of the devices that
                were removed, the second tuple contains the information of the new devices that were added.
        """
        current_devices, prev_devices = self.get_available_devices(), self.last_check_devices
        # Get the difference between the current devices and the previous ones
        removed_devices = {_id: _info for _id, _info in prev_devices.items() if _id not in current_devices}
        added_devices = {_id: _info for _id, _info in current_devices.items() if _id not in prev_devices}
        # Update the last checked devices to the current devices if requested
        if update_last_check_devices:
            self.last_check_devices = current_devices.copy()
        return removed_devices, added_devices

    @abstractmethod
    def get_available_devices(self) -> dict[str, dict[str, str | tuple[str, ...]]]:
        """
        Returns a dictionary of the currently available devices, where the key is the device ID and the value is a
        dictionary of the device's information.

        :return: dict[str, dict[str, str|tuple[str, ...]]. The key is the device ID, the value is a dictionary of the device's
                information.
        """
        raise NotImplementedError("This method must be implemented in the child class")

    def _apply_devices_filter(self, devices: dict[str, dict[str, str | tuple[str, ...]]]) -> dict[str, dict[str, str | tuple[str, ...]]]:
        """
        Filters the devices by the given filters. Only devices that match all the filters in any of the dictionaries
        will be returned.
        :param devices: dict[str, dict[str, str|tuple[str, ...]]]. The devices to filter. Returned by the
                `get_available_devices` method.
        :return: dict[str, dict[str, str|tuple[str, ...]]]. The input devices that matches any of the given filters
        """
        # Copy the dict to avoid allow iteration while modifying the original dict
        for device_id, device_info in devices.copy().items():
            # Iterate over each filter dict
            for filter_dict in self.filter_devices:
                if all(device_info[key] == value for key, value in filter_dict.items()):
                    break
            else:
                devices.pop(device_id)
        return devices

    def check_changes(self, on_connect: callable | None = None, on_disconnect: callable | None = None,
                      update_last_check_devices: bool = True) -> None:
        """
        Checks for changes in the USB devices. If a device is removed, the `on_disconnect` function will be called
        with the device ID as the first argument and the device information as the second argument. If a device is
        added, the `on_connect` function with the same arguments.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive
                two arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to
                receive two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])
        :param update_last_check_devices: bool. Whether to update the last checked devices to the current devices
        """
        removed_devices, added_devices = self.changes_from_last_check(update_last_check_devices=update_last_check_devices)
        if on_disconnect is not None:
            for device_id, device_info in removed_devices.items():
                on_disconnect(device_id, device_info)
        if on_connect is not None:
            for device_id, device_info in added_devices.items():
                on_connect(device_id, device_info)

    def start_monitoring(self, on_connect: callable|None = None, on_disconnect: callable|None = None,
                         check_every_seconds: int | float = _SECONDS_BETWEEN_CHECKS) -> None:
        """
        Starts monitoring the USB devices. This function will trigger a background thread that will check for changes
        in the USB devices every `check_every_seconds` seconds. If a device is removed, the `on_disconnect` function
        will be called with the device ID as the first argument and the device information as the second argument.
        If a device is added, the `on_connect` function with the same arguments.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive two
                arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to receive
                two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])

        :param check_every_seconds: int | float. The number of seconds to wait between each check for changes in the
                USB devices. Defaults to 0.5 seconds.
        """
        assert self._thread is None, "The USB monitor is already running"
        self._thread = threading.Thread(name="USB Monitor", target=self._monitor_changes,
                                        args=(on_connect, on_disconnect, check_every_seconds),
                                        daemon=True)
        self._thread.start()

    def _monitor_changes(self, on_connect: callable | None = None, on_disconnect: callable | None = None,
                             check_every_seconds: int | float = _SECONDS_BETWEEN_CHECKS) -> None:
        """
        Monitors the USB devices. This function should ALWAYS be called from a background thread.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive two
                arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to receive
                two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])
        :param check_every_seconds: int | float. The number of seconds to wait between each check for changes in the
                USB devices. Defaults to 0.5 seconds.
        """
        assert self._thread is not None, "The USB monitor is not running"
        if self._stop_thread.is_set():
            warn("USB monitor can not be started because it is already stopped. Call stop_monitoring() first",
                 RuntimeWarning)
        while not self._stop_thread.is_set():
            self.check_changes(on_connect=on_connect, on_disconnect=on_disconnect)
            self._stop_thread.wait(check_every_seconds)


    def stop_monitoring(self, warn_if_was_stopped: bool = True, warn_if_timeout: bool = True,
                        timeout=_THREAD_JOIN_TIMEOUT_SECONDS) -> None:
        """
        Stops monitoring the USB devices.
        :param warn_if_was_stopped: bool. Whether to warn if the USB monitor was already stopped.
        """
        if self._thread is not None:
            self._stop_thread.set()
            self._thread.join(timeout=timeout)
            if warn_if_timeout and self._thread.is_alive():
                warn(f"USB monitor thread did not stop in {timeout} seconds. "
                     f"It could still be running", RuntimeWarning)
            self._thread = None
        elif warn_if_was_stopped:
            warn("USB monitor can not be stopped because it is not running", RuntimeWarning)
        self._stop_thread.clear()

    def __del__(self):
        self.stop_monitoring(warn_if_was_stopped=False)


"""
USBMonitor: USBMonitor is an easy-to-use cross-platform library for USB device monitoring that simplifies
            tracking of connections, disconnections, and examination of connected device attributes on both
            Windows and Linux, freeing the user from platform-specific details or incompatibilities.

Author: Eric-Canas
Date: 24-03-2023
Email: eric@ericcanas.com
Github: https://github.com/Eric-Canas
"""

from __future__ import annotations
from .__platform_specific_detectors._constants import _SECONDS_BETWEEN_CHECKS, _THREAD_JOIN_TIMEOUT_SECONDS
from warnings import warn

import sys


class USBMonitor:
    def __init__(self, filter_devices: list[dict[str, str]] | tuple[dict[str, str]] | None = None):
        """
        Creates a new USBMonitor object. This object can be used to monitor USB devices connected to the system.

        :param filter_devices: list[dict[str, str]] | tuple[dict[str, str]] | None. A list of dictionaries containing
        the device information to filter the devices by. If None, no filtering will be done. The dictionaries
        must contain the same keys as the dictionaries returned by the `get_available_devices` method.
        For example, if you want to only monitor devices with ID_MODEL_ID = "A2B2" or "ABCD" you could pass
        filter_devices=({"ID_MODEL_ID": "A2B2"}, {"ID_MODEL_ID": "ABCD"}).
        """
        if sys.platform.startswith('linux'):
            from .__platform_specific_detectors._linux_usb_detector import _LinuxUSBDetector
            self.monitor = _LinuxUSBDetector(filter_devices=filter_devices)
        elif sys.platform.startswith('win'):
            from .__platform_specific_detectors._windows_usb_detector import _WindowsUSBDetector
            self.monitor = _WindowsUSBDetector(filter_devices=filter_devices)
        elif sys.platform.startswith('darwin'):
            from .__platform_specific_detectors._darwin_usb_detector import _DarwinUSBDetector
            self.monitor = _DarwinUSBDetector(filter_devices=filter_devices)
        else:
            raise NotImplementedError(f"Your OS is not supported: {sys.platform}")


    def get_available_devices(self) -> dict[str, dict[str, str | tuple[str, ...]]]:
        """
        Returns a dictionary of the currently available devices, where the key is the device ID and the value is a
        dictionary of the device's information. These keys IDs can be found at attributes.DEVICE_ATTRIBUTES. They
        do always correspond with the default Linux IDs (independently of the OS where the library is running).

        :return: dict[str, dict[str, str|tuple[str, ...]]]. The key is the device ID, the value is a dictionary of the device's
                information.
        """
        return self.monitor.get_available_devices()

    def check_changes(self, on_connect: callable | None = None, on_disconnect: callable | None = None,
                      update_last_check_devices: bool = True) -> None:
        """
        Checks for changes in the USB devices. If a device is removed, the `on_disconnect` function will be called
        with the device ID as the first argument and the device information as the second argument. If a device is
        added, the `on_connect` function with the same arguments.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive
                two arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to
                receive two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])
        :param update_last_check_devices: bool. Whether to update the last checked devices to the current devices
        """
        self.monitor.check_changes(on_connect=on_connect, on_disconnect=on_disconnect,
                                   update_last_check_devices=update_last_check_devices)

    def changes_from_last_check(self, update_last_check_devices: bool = True) -> tuple[dict[str, str], dict[str, str]]:
        """
        Returns a tuple of two tuples, the first containing the device IDs of the devices that were removed, the second
        containing the device IDs of the devices that were added.
        :param update_last_check_devices: bool. Whether to update the last checked devices to the current devices
        :return: tuple[dict[str, str], dict[str, str]]. The first tuple contains the information of the devices that
                were removed, the second tuple contains the information of the new devices that were added.
        """
        return self.monitor.changes_from_last_check(update_last_check_devices=update_last_check_devices)

    def start_monitoring(self, on_connect: callable|None = None, on_disconnect: callable|None = None,
                         check_every_seconds: int | float = _SECONDS_BETWEEN_CHECKS) -> None:
        """
        Starts monitoring the USB devices. This function will trigger a background thread that will check for changes
        in the USB devices every `check_every_seconds` seconds. If a device is removed, the `on_disconnect` function
        will be called with the device ID as the first argument and the device information as the second argument.
        If a device is added, the `on_connect` function with the same arguments.
        :param on_connect: callable | None. The function to call when a device is added. It is expected to receive two
                arguments, the device ID and the device information. on_connect(device_id: str, device_info: dict[str, str])
        :param on_disconnect: callable | None. The function to call when a device is removed. It is expected to receive
                two arguments, the device ID and the device information. on_disconnect(device_id: str, device_info: dict[str, str])
        :param check_every_seconds: int | float. The number of seconds to wait between each check for changes in the
                USB devices. Defaults to 0.5 seconds.
        """
        if on_connect is None and on_disconnect is None:
            warn("You are starting the monitor without any callback functions. This won't notice anything "
                 "when a device is connected or disconnected.")
        self.monitor.start_monitoring(on_connect=on_connect, on_disconnect=on_disconnect,
                                      check_every_seconds=check_every_seconds)

    def stop_monitoring(self, timeout=_THREAD_JOIN_TIMEOUT_SECONDS) -> None:
        """
        Stops monitoring the USB devices. This function will stop the background thread that was checking for changes
        in the USB devices.
        """
        self.monitor.stop_monitoring(timeout=timeout)

    # When requesting a function that does not exist, it will be redirected to the monitor
    def __getattr__(self, item):
        if hasattr(self.monitor, item):
            return getattr(self.monitor, item)
        else:
            raise AttributeError(f"USBMonitor has no attribute '{item}'")